"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Builder = exports.ExportFormat = exports.EnumFormat = void 0;
var tslib_1 = require("tslib");
var events_1 = require("events");
var path_1 = require("path");
var graphql_1 = require("graphql");
var chalk_1 = tslib_1.__importDefault(require("chalk"));
var fs_extra_1 = require("fs-extra");
var chokidar_1 = require("chokidar");
var graphql_config_1 = require("graphql-config");
var graphql_config_utilities_1 = require("graphql-config-utilities");
var graphql_tool_utilities_1 = require("graphql-tool-utilities");
var print_1 = require("./print");
var types_1 = require("./types");
Object.defineProperty(exports, "EnumFormat", { enumerable: true, get: function () { return types_1.EnumFormat; } });
Object.defineProperty(exports, "ExportFormat", { enumerable: true, get: function () { return types_1.ExportFormat; } });
var Builder = /** @class */ (function (_super) {
    tslib_1.__extends(Builder, _super);
    function Builder(_a) {
        var cwd = _a.cwd, options = tslib_1.__rest(_a, ["cwd"]);
        var _this = _super.call(this) || this;
        // projectName -> {filePath -> document}
        // NOTE: projectName can be undefined for nameless graphql-config projects
        _this.documentMapByProject = new Map();
        _this.watchers = [];
        _this.options = options;
        _this.config = options.config
            ? options.config
            : graphql_config_1.getGraphQLConfig(cwd ? path_1.resolve(cwd) : undefined);
        return _this;
    }
    Builder.prototype.once = function (event, handler) {
        return _super.prototype.once.call(this, event, handler);
    };
    Builder.prototype.on = function (event, handler) {
        return _super.prototype.on.call(this, event, handler);
    };
    Builder.prototype.emit = function (event) {
        var args = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            args[_i - 1] = arguments[_i];
        }
        return _super.prototype.emit.apply(this, tslib_1.__spread([event], args));
    };
    Builder.prototype.run = function (_a) {
        var _b = (_a === void 0 ? {} : _a).watch, watchGlobs = _b === void 0 ? false : _b;
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var schemaPaths, error_1, error_2;
            var _c;
            var _this = this;
            return tslib_1.__generator(this, function (_d) {
                switch (_d.label) {
                    case 0:
                        try {
                            schemaPaths = graphql_config_utilities_1.getGraphQLSchemaPaths(this.config);
                        }
                        catch (error) {
                            this.emit('error', error);
                            return [2 /*return*/];
                        }
                        if (!watchGlobs) return [3 /*break*/, 2];
                        (_c = this.watchers).push.apply(_c, tslib_1.__spread(this.setupDocumentWatchers().concat(this.setupSchemaWatcher())));
                        // wait for all watchers to be ready
                        return [4 /*yield*/, Promise.all(this.watchers.map(function (watcher) {
                                return new Promise(function (resolve) { return watcher.on('ready', function () { return resolve(); }); });
                            }))];
                    case 1:
                        // wait for all watchers to be ready
                        _d.sent();
                        _d.label = 2;
                    case 2:
                        _d.trys.push([2, 4, , 5]);
                        this.emit('start:schema');
                        return [4 /*yield*/, Promise.all(schemaPaths.map(function (schemaPath) { return _this.generateSchemaTypes(schemaPath); }))];
                    case 3:
                        _d.sent();
                        this.emit('end:schema');
                        return [3 /*break*/, 5];
                    case 4:
                        error_1 = _d.sent();
                        this.emit('error', error_1);
                        return [2 /*return*/];
                    case 5:
                        _d.trys.push([5, 7, , 8]);
                        return [4 /*yield*/, Promise.all(graphql_config_utilities_1.getGraphQLProjects(this.config).map(function (projectConfig) {
                                return _this.updateDocumentsForProject(projectConfig);
                            }))];
                    case 6:
                        _d.sent();
                        return [3 /*break*/, 8];
                    case 7:
                        error_2 = _d.sent();
                        this.emit('error', error_2);
                        return [2 /*return*/];
                    case 8: return [4 /*yield*/, this.generateDocumentTypes()];
                    case 9:
                        _d.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    Builder.prototype.stop = function () {
        this.watchers.forEach(function (watcher) {
            watcher.close();
        });
        this.watchers.length = 0;
    };
    Builder.prototype.setupDocumentWatchers = function () {
        var _this = this;
        var update = function (filePath, projectConfig) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var error_3;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        _a.trys.push([0, 2, , 3]);
                        return [4 /*yield*/, this.updateDocumentForFile(filePath, projectConfig)];
                    case 1:
                        _a.sent();
                        return [3 /*break*/, 3];
                    case 2:
                        error_3 = _a.sent();
                        this.emit('error', error_3);
                        return [2 /*return*/];
                    case 3: return [4 /*yield*/, this.generateDocumentTypes()];
                    case 4:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        }); };
        return graphql_config_utilities_1.getGraphQLProjects(this.config)
            .filter(function (_a) {
            var includes = _a.includes;
            return includes.length > 0;
        })
            .map(function (projectConfig) {
            return chokidar_1.watch(projectConfig.includes.map(function (include) {
                return graphql_config_utilities_1.resolvePathRelativeToConfig(projectConfig, include);
            }), {
                ignored: projectConfig.excludes.map(function (exclude) {
                    return graphql_config_utilities_1.resolvePathRelativeToConfig(projectConfig, exclude);
                }),
                ignoreInitial: true,
            })
                .on('add', function (filePath) { return update(filePath, projectConfig); })
                .on('change', function (filePath) { return update(filePath, projectConfig); })
                .on('unlink', function (filePath) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                var documents;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            documents = this.documentMapByProject.get(projectConfig.projectName);
                            if (documents) {
                                documents.delete(filePath);
                            }
                            return [4 /*yield*/, this.generateDocumentTypes()];
                        case 1:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
        });
    };
    Builder.prototype.setupSchemaWatcher = function () {
        var _this = this;
        var update = function (schemaPath) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            var error_4;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        _a.trys.push([0, 3, , 4]);
                        this.emit('start:schema');
                        return [4 /*yield*/, this.generateSchemaTypes(schemaPath)];
                    case 1:
                        _a.sent();
                        this.emit('end:schema');
                        return [4 /*yield*/, this.generateDocumentTypes()];
                    case 2:
                        _a.sent();
                        return [3 /*break*/, 4];
                    case 3:
                        error_4 = _a.sent();
                        return [3 /*break*/, 4];
                    case 4: return [2 /*return*/];
                }
            });
        }); };
        return chokidar_1.watch(graphql_config_utilities_1.getGraphQLSchemaPaths(this.config), { ignoreInitial: true }).on('change', update);
    };
    Builder.prototype.generateSchemaTypes = function (schemaPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var projectConfig, schemaTypesPath, definitions;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        projectConfig = graphql_config_utilities_1.getGraphQLProjectForSchemaPath(this.config, schemaPath);
                        schemaTypesPath = getSchemaTypesPath(projectConfig, this.options);
                        definitions = print_1.generateSchemaTypes(projectConfig.getSchema(), this.options);
                        return [4 /*yield*/, fs_extra_1.mkdirp(schemaTypesPath)];
                    case 1:
                        _a.sent();
                        return [4 /*yield*/, Promise.all(Array.from(definitions.entries()).map(function (_a) {
                                var _b = tslib_1.__read(_a, 2), fileName = _b[0], definition = _b[1];
                                return fs_extra_1.writeFile(path_1.join(schemaTypesPath, fileName), definition);
                            }))];
                    case 2:
                        _a.sent();
                        this.emit('build:schema', {
                            schemaPath: schemaPath,
                            schemaTypesPath: schemaTypesPath,
                        });
                        return [2 /*return*/];
                }
            });
        });
    };
    Builder.prototype.generateDocumentTypes = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        this.emit('start:docs');
                        this.checkForDuplicateOperations();
                        this.checkForDuplicateFragments();
                        return [4 /*yield*/, Promise.all(Array.from(this.documentMapByProject.entries()).map(function (_a) {
                                var _b = tslib_1.__read(_a, 2), projectName = _b[0], documents = _b[1];
                                return _this.generateDocumentTypesForProject(_this.config.getProjectConfig(projectName), documents);
                            }))];
                    case 1:
                        _a.sent();
                        this.emit('end:docs');
                        return [2 /*return*/];
                }
            });
        });
    };
    Builder.prototype.checkForDuplicateOperations = function () {
        var _this = this;
        getDuplicateOperations(this.documentMapByProject).forEach(function (_a) {
            var projectName = _a.projectName, duplicates = _a.duplicates;
            if (duplicates.length) {
                duplicates.forEach(function (_a) {
                    var operationName = _a.operationName, filePaths = _a.filePaths;
                    var message = "GraphQL operations must have a unique name. The operation " + chalk_1.default.bold(operationName) + " is declared in:\n " + filePaths.sort().join('\n ') + (projectName ? " (" + chalk_1.default.bold(projectName) + ")" : '');
                    _this.emit('error', new Error(message));
                });
            }
        });
    };
    Builder.prototype.checkForDuplicateFragments = function () {
        var _this = this;
        getDuplicateFragments(this.documentMapByProject).forEach(function (_a) {
            var projectName = _a.projectName, duplicates = _a.duplicates;
            if (duplicates.length) {
                duplicates.forEach(function (_a) {
                    var fragmentName = _a.fragmentName, filePaths = _a.filePaths;
                    var message = "GraphQL fragments must have a unique name. The fragment " + chalk_1.default.bold(fragmentName) + " is declared in:\n " + filePaths.sort().join('\n ') + (projectName ? " (" + chalk_1.default.bold(projectName) + ")" : '');
                    _this.emit('error', new Error(message));
                });
            }
        });
    };
    Builder.prototype.generateDocumentTypesForProject = function (projectConfig, documents) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var ast, error_5;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        try {
                            ast = graphql_tool_utilities_1.compile(projectConfig.getSchema(), graphql_1.concatAST(Array.from(documents.values())));
                        }
                        catch (error) {
                            this.emit('error', error);
                            return [2 /*return*/];
                        }
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 3, , 4]);
                        return [4 /*yield*/, Promise.all(Array.from(groupOperationsAndFragmentsByFile(ast).values()).map(function (file) {
                                return _this.writeDocumentFile(file, ast, projectConfig);
                            }))];
                    case 2:
                        _a.sent();
                        return [3 /*break*/, 4];
                    case 3:
                        error_5 = _a.sent();
                        return [3 /*break*/, 4];
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    Builder.prototype.writeDocumentFile = function (file, ast, projectConfig) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var definitionPath;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        definitionPath = file.path + ".d.ts";
                        return [4 /*yield*/, fs_extra_1.writeFile(definitionPath, this.getDocumentDefinition(file, ast, projectConfig))];
                    case 1:
                        _a.sent();
                        this.emit('build:docs', {
                            documentPath: file.path,
                            definitionPath: definitionPath,
                            operation: file.operation,
                            fragments: file.fragments,
                        });
                        return [2 /*return*/];
                }
            });
        });
    };
    Builder.prototype.getDocumentDefinition = function (file, ast, projectConfig) {
        try {
            return print_1.printDocument(file, ast, {
                enumFormat: this.options.enumFormat,
                exportFormat: this.options.exportFormat,
                addTypename: this.options.addTypename,
                schemaTypesPath: getSchemaTypesPath(projectConfig, this.options),
            });
        }
        catch (_a) {
            var message = _a.message;
            var error = new Error("Error in " + file.path + ": " + message[0].toLowerCase() + message.slice(1));
            this.emit('error', error);
            throw error;
        }
    };
    Builder.prototype.updateDocumentsForProject = function (projectConfig) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var filePaths;
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, graphql_config_utilities_1.getGraphQLProjectIncludedFilePaths(projectConfig)];
                    case 1:
                        filePaths = _a.sent();
                        return [2 /*return*/, Promise.all(filePaths.map(function (filePath) {
                                return _this.updateDocumentForFile(filePath, projectConfig);
                            }))];
                }
            });
        });
    };
    Builder.prototype.updateDocumentForFile = function (filePath, projectConfig) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var contents;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, fs_extra_1.readFile(filePath, 'utf8')];
                    case 1:
                        contents = _a.sent();
                        return [2 /*return*/, this.setDocumentForFilePath(filePath, projectConfig, contents)];
                }
            });
        });
    };
    Builder.prototype.setDocumentForFilePath = function (filePath, projectConfig, contents) {
        var documents = this.documentMapByProject.get(projectConfig.projectName);
        if (!documents) {
            documents = new Map();
            this.documentMapByProject.set(projectConfig.projectName, documents);
        }
        if (contents.trim().length === 0) {
            return undefined;
        }
        var document = graphql_1.parse(new graphql_1.Source(contents, filePath));
        documents.set(filePath, document);
        return document;
    };
    return Builder;
}(events_1.EventEmitter));
exports.Builder = Builder;
function getSchemaTypesPath(projectConfig, options) {
    if (typeof projectConfig.extensions.schemaTypesPath === 'string') {
        return graphql_config_utilities_1.resolvePathRelativeToConfig(projectConfig, projectConfig.extensions.schemaTypesPath);
    }
    return graphql_config_utilities_1.resolvePathRelativeToConfig(projectConfig, path_1.join(options.schemaTypesPath, (projectConfig.projectName ? projectConfig.projectName + "-" : '') + "types"));
}
function groupOperationsAndFragmentsByFile(_a) {
    var operations = _a.operations, fragments = _a.fragments;
    return Object.values(operations)
        .concat(Object.values(fragments))
        .reduce(function (map, item) {
        if (!item.filePath) {
            return map;
        }
        var file = map.get(item.filePath);
        if (!file) {
            file = {
                path: item.filePath,
                operation: undefined,
                fragments: [],
            };
            map.set(item.filePath, file);
        }
        if (graphql_tool_utilities_1.isOperation(item)) {
            file.operation = item;
        }
        else {
            file.fragments.push(item);
        }
        return map;
    }, new Map());
}
function getDuplicateOperations(documentsMapByProject) {
    return Array.from(documentsMapByProject.entries()).map(function (_a) {
        var _b = tslib_1.__read(_a, 2), projectName = _b[0], documents = _b[1];
        return {
            projectName: projectName,
            duplicates: getDuplicateProjectOperations(documents),
        };
    });
}
function getDuplicateFragments(documentsMapByProject) {
    return Array.from(documentsMapByProject.entries()).map(function (_a) {
        var _b = tslib_1.__read(_a, 2), projectName = _b[0], documents = _b[1];
        return {
            projectName: projectName,
            duplicates: getDuplicateProjectFragments(documents),
        };
    });
}
function getDuplicateProjectOperations(documents) {
    var operations = new Map();
    Array.from(documents.entries()).forEach(function (_a) {
        var _b = tslib_1.__read(_a, 2), filePath = _b[0], document = _b[1];
        document.definitions.filter(isOperationDefinition).forEach(function (definition) {
            var name = definition.name;
            if (name && name.value) {
                var map = operations.get(name.value);
                if (map) {
                    map.add(filePath);
                }
                else {
                    operations.set(name.value, new Set([filePath]));
                }
            }
        });
    });
    return Array.from(operations.entries())
        .filter(function (_a) {
        var _b = tslib_1.__read(_a, 2), filePaths = _b[1];
        return filePaths.size > 1;
    })
        .map(function (_a) {
        var _b = tslib_1.__read(_a, 2), operationName = _b[0], filePath = _b[1];
        return { operationName: operationName, filePaths: Array.from(filePath) };
    });
}
function getDuplicateProjectFragments(documents) {
    var fragments = new Map();
    Array.from(documents.entries()).forEach(function (_a) {
        var _b = tslib_1.__read(_a, 2), filePath = _b[0], document = _b[1];
        document.definitions.filter(isFragmentDefinition).forEach(function (definition) {
            var name = definition.name;
            if (name && name.value) {
                var map = fragments.get(name.value);
                if (map) {
                    map.add(filePath);
                }
                else {
                    fragments.set(name.value, new Set([filePath]));
                }
            }
        });
    });
    return Array.from(fragments.entries())
        .filter(function (_a) {
        var _b = tslib_1.__read(_a, 2), filePaths = _b[1];
        return filePaths.size > 1;
    })
        .map(function (_a) {
        var _b = tslib_1.__read(_a, 2), fragmentName = _b[0], filePath = _b[1];
        return { fragmentName: fragmentName, filePaths: Array.from(filePath) };
    });
}
function isOperationDefinition(definition) {
    return definition.kind === 'OperationDefinition';
}
function isFragmentDefinition(definition) {
    return definition.kind === 'FragmentDefinition';
}
