"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OperationContext = exports.FileContext = void 0;
var tslib_1 = require("tslib");
var path_1 = require("path");
var t = tslib_1.__importStar(require("@babel/types"));
var upper_case_first_1 = require("upper-case-first");
var graphql_tool_utilities_1 = require("graphql-tool-utilities");
var FileContext = /** @class */ (function () {
    function FileContext(path, options) {
        this.path = path;
        this.options = options;
        this.importedTypes = new Set();
    }
    Object.defineProperty(FileContext.prototype, "schemaImports", {
        get: function () {
            var _a = this, path = _a.path, importedTypes = _a.importedTypes, schemaTypesPath = _a.options.schemaTypesPath;
            return importedTypes.size > 0
                ? t.importDeclaration(tslib_1.__spread(importedTypes).map(function (type) {
                    return t.importSpecifier(t.identifier(type), t.identifier(type));
                }), t.stringLiteral(importPath(path, schemaTypesPath)))
                : null;
        },
        enumerable: false,
        configurable: true
    });
    FileContext.prototype.import = function (type) {
        this.importedTypes.add(type);
    };
    return FileContext;
}());
exports.FileContext = FileContext;
var OperationContext = /** @class */ (function () {
    function OperationContext(operation, ast, options, file) {
        this.operation = operation;
        this.ast = ast;
        this.options = options;
        this.file = file;
        this.exportedTypes = [];
    }
    Object.defineProperty(OperationContext.prototype, "typeName", {
        get: function () {
            var typeName;
            if (graphql_tool_utilities_1.isOperation(this.operation)) {
                var _a = this.operation, operationName = _a.operationName, operationType = _a.operationType;
                typeName = "" + upper_case_first_1.upperCaseFirst(operationName) + upper_case_first_1.upperCaseFirst(operationType) + "Data";
            }
            else {
                var fragmentName = this.operation.fragmentName;
                typeName = upper_case_first_1.upperCaseFirst(fragmentName) + "FragmentData";
            }
            return this.options.partial
                ? typeName.replace(/Data$/, 'PartialData')
                : typeName;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(OperationContext.prototype, "namespace", {
        get: function () {
            var _a = this, exported = _a.exported, typeName = _a.typeName;
            return exported.length > 0
                ? t.tsModuleDeclaration(t.identifier(typeName), t.tsModuleBlock(exported.map(function (type) { return t.exportNamedDeclaration(type, []); })))
                : null;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(OperationContext.prototype, "exported", {
        get: function () {
            return this.exportedTypes;
        },
        enumerable: false,
        configurable: true
    });
    OperationContext.prototype.export = function (type) {
        this.exportedTypes.push(type);
        return t.tsTypeReference(t.tsQualifiedName(t.identifier(this.typeName), t.identifier(type.id.name)));
    };
    return OperationContext;
}());
exports.OperationContext = OperationContext;
function importPath(from, to) {
    var relativePath = path_1.relative(path_1.dirname(from), to);
    var normalizedPath = relativePath.startsWith('..')
        ? relativePath
        : "./" + relativePath;
    return normalizedPath.replace(/\.ts$/, '');
}
