"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.printDocument = void 0;
var tslib_1 = require("tslib");
var t = tslib_1.__importStar(require("@babel/types"));
var graphql_tool_utilities_1 = require("graphql-tool-utilities");
var types_1 = require("../../types");
var context_1 = require("./context");
var utilities_1 = require("./utilities");
var language_1 = require("./language");
var generate = require('@babel/generator').default;
function printDocument(_a, ast, options) {
    var path = _a.path, operation = _a.operation, fragments = _a.fragments;
    var file = new context_1.FileContext(path, options);
    if (operation == null) {
        var fileBody_1 = fragments.reduce(function (statements, fragment) {
            var context = new context_1.OperationContext(fragment, ast, options, file);
            var body = language_1.tsInterfaceBodyForObjectField(fragment, fragment.typeCondition, new utilities_1.ObjectStack(fragment.typeCondition), context);
            var namespace = context.namespace;
            return tslib_1.__spread(statements, (namespace ? [t.exportNamedDeclaration(namespace, [])] : []), [
                t.exportNamedDeclaration(t.tsInterfaceDeclaration(t.identifier(context.typeName), null, null, body), []),
            ]);
        }, []);
        var schemaImports_1 = file.schemaImports;
        if (schemaImports_1) {
            fileBody_1.unshift(schemaImports_1);
        }
        return generate(t.file(t.program(fileBody_1), [], [])).code;
    }
    var context = new context_1.OperationContext(operation, ast, options, file);
    var partialContext = new context_1.OperationContext(operation, ast, tslib_1.__assign(tslib_1.__assign({}, options), { partial: true }), file);
    var rootType;
    if (operation.operationType === graphql_tool_utilities_1.OperationType.Query) {
        rootType = ast.schema.getQueryType();
    }
    else if (operation.operationType === graphql_tool_utilities_1.OperationType.Mutation) {
        rootType = ast.schema.getMutationType();
    }
    else {
        rootType = ast.schema.getSubscriptionType();
    }
    var variables = operation.variables.filter(graphql_tool_utilities_1.isTypedVariable).length > 0
        ? context.export(language_1.variablesInterface(operation.variables, context))
        : null;
    var operationInterface = t.tsInterfaceDeclaration(t.identifier(context.typeName), null, null, language_1.tsInterfaceBodyForObjectField(operation, rootType, new utilities_1.ObjectStack(rootType), context));
    var operationPartialInterface = t.tsInterfaceDeclaration(t.identifier(partialContext.typeName), null, null, language_1.tsInterfaceBodyForObjectField(operation, rootType, new utilities_1.ObjectStack(rootType), partialContext));
    var schemaImports = file.schemaImports;
    var namespace = context.namespace;
    var partialNamespace = partialContext.namespace;
    var _b = options.exportFormat, exportFormat = _b === void 0 ? types_1.ExportFormat.Document : _b;
    var documentType = exportFormat === types_1.ExportFormat.Document ? 'DocumentNode' : 'SimpleDocument';
    var documentNodeImport = t.importDeclaration([t.importSpecifier(t.identifier(documentType), t.identifier(documentType))], t.stringLiteral('graphql-typed'));
    var documentNodeDeclaratorIdentifier = t.identifier('document');
    documentNodeDeclaratorIdentifier.typeAnnotation = t.tsTypeAnnotation(t.tsTypeReference(t.identifier(documentType), t.tsTypeParameterInstantiation([
        t.tsTypeReference(t.identifier(context.typeName)),
        variables || t.tsNeverKeyword(),
        t.tsTypeReference(t.identifier(partialContext.typeName)),
    ])));
    var documentNodeDeclaration = t.variableDeclaration('const', [
        t.variableDeclarator(documentNodeDeclaratorIdentifier),
    ]);
    documentNodeDeclaration.declare = true;
    var documentNodeExport = t.exportDefaultDeclaration(t.identifier('document'));
    var fileBody = [documentNodeImport];
    if (schemaImports) {
        fileBody.push(schemaImports);
    }
    if (partialNamespace) {
        fileBody.push(t.exportNamedDeclaration(partialNamespace, []));
    }
    fileBody.push(t.exportNamedDeclaration(operationPartialInterface, []));
    if (namespace) {
        fileBody.push(t.exportNamedDeclaration(namespace, []));
    }
    fileBody.push(t.exportNamedDeclaration(operationInterface, []), documentNodeDeclaration, documentNodeExport);
    return generate(t.file(t.program(fileBody), [], [])).code;
}
exports.printDocument = printDocument;
