"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.variablesInterface = exports.tsInterfaceBodyForObjectField = void 0;
var tslib_1 = require("tslib");
var t = tslib_1.__importStar(require("@babel/types"));
var graphql_1 = require("graphql");
var graphql_tool_utilities_1 = require("graphql-tool-utilities");
var utilities_1 = require("../utilities");
function tsInterfaceBodyForObjectField(_a, graphQLType, stack, context, requiresTypename) {
    var _b = _a.fields, fields = _b === void 0 ? [] : _b;
    if (requiresTypename === void 0) { requiresTypename = false; }
    var schema = context.ast.schema;
    var isRootType = !Array.isArray(graphQLType) &&
        (graphQLType === schema.getQueryType() ||
            graphQLType === schema.getMutationType() ||
            graphQLType === schema.getSubscriptionType());
    var uniqueFields = fields.filter(function (field) {
        if (stack.hasSeenField(field)) {
            return false;
        }
        stack.sawField(field);
        return true;
    });
    var typenameField = {
        fieldName: '__typename',
        responseName: '__typename',
        type: new graphql_1.GraphQLNonNull(graphql_1.GraphQLString),
        isConditional: false,
    };
    var typename = ((context.options.addTypename && !isRootType) || requiresTypename) &&
        !stack.hasSeenField(typenameField)
        ? tsPropertyForField(typenameField, graphQLType, stack, context, requiresTypename)
        : null;
    var body = uniqueFields.map(function (field) {
        return tsPropertyForField(field, graphQLType, stack, context, requiresTypename);
    });
    return t.tsInterfaceBody(typename ? tslib_1.__spread([typename], body) : body);
}
exports.tsInterfaceBodyForObjectField = tsInterfaceBodyForObjectField;
function tsTypeForInlineFragment(inlineFragment, _graphQLType, stack, context, requiresTypename) {
    if (requiresTypename === void 0) { requiresTypename = false; }
    var typeCondition = inlineFragment.typeCondition;
    var interfaceDeclaration = t.tsInterfaceDeclaration(t.identifier(stack.name), null, null, tsInterfaceBodyForObjectField(inlineFragment, typeCondition, stack, context, requiresTypename));
    return context.export(interfaceDeclaration);
}
function tsTypeForObjectField(field, graphQLType, stack, context) {
    var _a = field.inlineFragments, inlineFragments = _a === void 0 ? [] : _a;
    var possibleTypes = graphql_1.isInterfaceType(graphQLType) || graphql_1.isUnionType(graphQLType)
        ? context.ast.schema.getPossibleTypes(graphQLType)
        : [];
    if (inlineFragments.length) {
        var fragmentTypes = tslib_1.__spread(inlineFragments).map(function (inlineFragment) {
            return tsTypeForInlineFragment(inlineFragment, graphQLType, stack.fragment(inlineFragment.typeCondition), context, context.options.partial);
        });
        var typesCoveredByInlineFragments_1 = new Set(tslib_1.__spread(inlineFragments).reduce(function (types, inlineFragment) { return tslib_1.__spread(types, inlineFragment.possibleTypes); }, []));
        var missingPossibleTypes = possibleTypes.filter(function (possibleType) {
            return !typesCoveredByInlineFragments_1.has(possibleType);
        });
        var otherStack = stack.fragment();
        var otherTypeInterface = t.tsInterfaceDeclaration(t.identifier(otherStack.name), null, null, tsInterfaceBodyForObjectField(field, missingPossibleTypes, otherStack, context, context.options.partial));
        var otherType = context.export(otherTypeInterface);
        return t.tsUnionType(tslib_1.__spread(fragmentTypes, [otherType]));
    }
    else if (possibleTypes.length === 1) {
        // When we have an interface or union type, but no inline fragments, it
        // means that there is only one conforming type for the union/ interface.
        // Here, we construct a "nothing" type that stands in for future additions
        // to the membership of the union/ interface.
        var otherStack = stack.fragment();
        var otherTypeInterface = t.tsInterfaceDeclaration(t.identifier(otherStack.name), null, null, tsInterfaceBodyForObjectField({ fields: [] }, [], otherStack, context, context.options.partial));
        var interfaceStack = stack.fragment(possibleTypes[0]);
        var interfaceDeclaration_1 = t.tsInterfaceDeclaration(t.identifier(interfaceStack.name), null, null, tsInterfaceBodyForObjectField(field, graphQLType, interfaceStack, context));
        return t.tsUnionType([
            context.export(interfaceDeclaration_1),
            context.export(otherTypeInterface),
        ]);
    }
    var interfaceDeclaration = t.tsInterfaceDeclaration(t.identifier(stack.name), null, null, tsInterfaceBodyForObjectField(field, graphQLType, stack, context));
    return context.export(interfaceDeclaration);
}
function tsTypenameForGraphQLType(type) {
    return t.tsLiteralType(t.stringLiteral(type.name));
}
function tsPropertyForField(field, parentType, stack, context, isRequiredTypename) {
    if (isRequiredTypename === void 0) { isRequiredTypename = false; }
    if (field.fieldName === '__typename' && parentType) {
        var optional = !isRequiredTypename &&
            (context.options.partial ||
                field.isConditional ||
                !graphql_1.isNonNullType(field.type));
        var parentTypes = Array.isArray(parentType) ? parentType : [parentType];
        var allPossibleTypes = parentTypes.reduce(function (all, type) { return tslib_1.__spread(all, (graphql_1.isAbstractType(type)
            ? context.ast.schema.getPossibleTypes(type)
            : [type])); }, []);
        var typename = void 0;
        if (allPossibleTypes.length === 0) {
            typename = t.tsLiteralType(t.stringLiteral(''));
        }
        else if (allPossibleTypes.length > 1) {
            typename = t.tsUnionType(allPossibleTypes.map(tsTypenameForGraphQLType));
        }
        else {
            typename = tsTypenameForGraphQLType(allPossibleTypes[0]);
        }
        var typenameProperty = t.tsPropertySignature(t.identifier(field.responseName), optional
            ? t.tsTypeAnnotation(t.tsUnionType([typename, t.tsNullKeyword()]))
            : t.tsTypeAnnotation(typename));
        typenameProperty.optional = optional;
        return typenameProperty;
    }
    var property = t.tsPropertySignature(t.identifier(field.responseName), t.tsTypeAnnotation(tsTypeForGraphQLType(field.type, field, stack, context)));
    property.optional =
        context.options.partial ||
            field.isConditional ||
            !graphql_1.isNonNullType(field.type);
    return property;
}
function tsTypeForGraphQLType(graphQLType, field, stack, context) {
    var type;
    var forceNullable = context.options.partial ||
        (field.isConditional && graphQLType === field.type);
    var isNonNull = !forceNullable && graphql_1.isNonNullType(graphQLType);
    var unwrappedGraphQLType = graphql_1.isNonNullType(graphQLType)
        ? graphQLType.ofType
        : graphQLType;
    if (graphql_1.isScalarType(unwrappedGraphQLType)) {
        if (Object.prototype.hasOwnProperty.call(utilities_1.scalarTypeMap, unwrappedGraphQLType.name)) {
            type = utilities_1.scalarTypeMap[unwrappedGraphQLType.name];
        }
        else {
            context.file.import(unwrappedGraphQLType.name);
            type = t.tsTypeReference(t.identifier(unwrappedGraphQLType.name));
        }
    }
    else if (graphql_1.isEnumType(unwrappedGraphQLType)) {
        context.file.import(unwrappedGraphQLType.name);
        type = t.tsTypeReference(t.identifier(unwrappedGraphQLType.name));
    }
    else if (graphql_1.isListType(unwrappedGraphQLType)) {
        var ofType = unwrappedGraphQLType.ofType;
        var arrayType = tsTypeForGraphQLType(ofType, field, stack, context);
        type = t.tsArrayType(t.isTSUnionType(arrayType) ? t.tsParenthesizedType(arrayType) : arrayType);
    }
    else if (graphql_1.isObjectType(unwrappedGraphQLType) ||
        graphql_1.isInterfaceType(unwrappedGraphQLType) ||
        graphql_1.isUnionType(unwrappedGraphQLType)) {
        type = tsTypeForObjectField(field, unwrappedGraphQLType, stack.nested(field, unwrappedGraphQLType), context);
    }
    else {
        type = t.tsAnyKeyword();
    }
    return isNonNull ? type : t.tsUnionType([type, t.tsNullKeyword()]);
}
function variablesInterface(variables, context) {
    return t.tsInterfaceDeclaration(t.identifier('Variables'), null, null, t.tsInterfaceBody(variables
        .filter(graphql_tool_utilities_1.isTypedVariable)
        .map(function (variable) { return tsPropertyForVariable(variable, context); })));
}
exports.variablesInterface = variablesInterface;
function tsPropertyForVariable(_a, context) {
    var name = _a.name, type = _a.type;
    var property = t.tsPropertySignature(t.identifier(name), t.tsTypeAnnotation(tsTypeForGraphQLInputType(type, context)));
    property.optional = !graphql_1.isNonNullType(type);
    return property;
}
function tsTypeForGraphQLInputType(graphQLType, context) {
    var type;
    var unwrappedGraphQLType = graphql_1.isNonNullType(graphQLType)
        ? graphQLType.ofType
        : graphQLType;
    if (graphql_1.isScalarType(unwrappedGraphQLType)) {
        if (Object.prototype.hasOwnProperty.call(utilities_1.scalarTypeMap, unwrappedGraphQLType.name)) {
            type = utilities_1.scalarTypeMap[unwrappedGraphQLType.name];
        }
        else {
            context.file.import(unwrappedGraphQLType.name);
            type = t.tsTypeReference(t.identifier(unwrappedGraphQLType.name));
        }
    }
    else if (graphql_1.isEnumType(unwrappedGraphQLType) ||
        graphql_1.isInputObjectType(unwrappedGraphQLType)) {
        context.file.import(unwrappedGraphQLType.name);
        type = t.tsTypeReference(t.identifier(unwrappedGraphQLType.name));
    }
    else {
        var ofType = unwrappedGraphQLType.ofType;
        var arrayType = tsTypeForGraphQLInputType(ofType, context);
        type = t.tsArrayType(graphql_1.isNonNullType(ofType) ? arrayType : t.tsParenthesizedType(arrayType));
    }
    return graphql_1.isNonNullType(graphQLType)
        ? type
        : t.tsUnionType([type, t.tsNullKeyword()]);
}
