"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateSchemaTypes = void 0;
var tslib_1 = require("tslib");
var t = tslib_1.__importStar(require("@babel/types"));
var change_case_1 = require("change-case");
var graphql_1 = require("graphql");
var utilities_1 = require("../utilities");
var types_1 = require("../../types");
var generate = require('@babel/generator').default;
function generateSchemaTypes(schema, options) {
    var e_1, _a;
    if (options === void 0) { options = {}; }
    var importFileBody = [];
    var exportFileBody = [];
    var definitions = new Map();
    try {
        for (var _b = tslib_1.__values(Object.values(schema.getTypeMap())), _c = _b.next(); !_c.done; _c = _b.next()) {
            var type = _c.value;
            if (!graphql_1.isInputType(type) || type.name.startsWith('__')) {
                continue;
            }
            if (graphql_1.isScalarType(type) &&
                Object.prototype.hasOwnProperty.call(utilities_1.scalarTypeMap, type.name)) {
                continue;
            }
            if (graphql_1.isEnumType(type)) {
                var enumType = tsEnumForType(type, options);
                definitions.set(enumType.id.name + ".ts", generate(t.file(t.program([t.exportNamedDeclaration(enumType, [])]), [], [])).code);
                importFileBody.unshift(t.importDeclaration([t.importSpecifier(enumType.id, enumType.id)], t.stringLiteral("./" + enumType.id.name)));
                exportFileBody.unshift(t.exportNamedDeclaration(null, [
                    t.exportSpecifier(enumType.id, enumType.id),
                ]));
            }
            else if (graphql_1.isScalarType(type)) {
                var _d = options.customScalars, customScalars = _d === void 0 ? {} : _d;
                var customScalarDefinition = customScalars[type.name];
                var scalarType = tsScalarForType(type, customScalarDefinition);
                if (customScalarDefinition && customScalarDefinition.package) {
                    importFileBody.unshift(t.importDeclaration([
                        t.importSpecifier(t.identifier(makeCustomScalarImportNameSafe(customScalarDefinition.name, type.name)), t.identifier(customScalarDefinition.name)),
                    ], t.stringLiteral(customScalarDefinition.package)));
                    exportFileBody.unshift(t.exportNamedDeclaration(scalarType, []));
                }
                else {
                    exportFileBody.push(t.exportNamedDeclaration(scalarType, []));
                }
            }
            else {
                exportFileBody.push(t.exportNamedDeclaration(tsInputObjectForType(type), []));
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
        }
        finally { if (e_1) throw e_1.error; }
    }
    // A blank file is ambiguous - its not clear if it is a script or a module.
    // If the file would be blank then give it an empty `export {}` to make in
    // unambiguously an es module file.
    // This ensures the generated files passes TypeScript type checking in
    // "isolatedModules" mode.
    if (exportFileBody.length === 0) {
        exportFileBody.push(t.exportNamedDeclaration(null, []));
    }
    return definitions.set('index.ts', generate(t.file(t.program(importFileBody.concat(exportFileBody)), [], []))
        .code);
}
exports.generateSchemaTypes = generateSchemaTypes;
function tsTypeForInputType(type) {
    var unwrappedType = graphql_1.isNonNullType(type) ? type.ofType : type;
    var tsType;
    if (graphql_1.isListType(unwrappedType)) {
        var tsTypeOfContainedType = tsTypeForInputType(unwrappedType.ofType);
        tsType = t.tsArrayType(t.isTSUnionType(tsTypeOfContainedType)
            ? t.tsParenthesizedType(tsTypeOfContainedType)
            : tsTypeOfContainedType);
    }
    else if (graphql_1.isScalarType(unwrappedType)) {
        tsType =
            utilities_1.scalarTypeMap[unwrappedType.name] ||
                t.tsTypeReference(t.identifier(unwrappedType.name));
    }
    else {
        tsType = t.tsTypeReference(t.identifier(unwrappedType.name));
    }
    return graphql_1.isNonNullType(type)
        ? tsType
        : t.tsUnionType([tsType, t.tsNullKeyword()]);
}
function tsInputObjectForType(type) {
    var fields = Object.entries(type.getFields()).map(function (_a) {
        var _b = tslib_1.__read(_a, 2), name = _b[0], field = _b[1];
        var property = t.tsPropertySignature(t.identifier(name), t.tsTypeAnnotation(tsTypeForInputType(field.type)));
        property.optional = !graphql_1.isNonNullType(field.type);
        return property;
    });
    return t.tsInterfaceDeclaration(t.identifier(type.name), null, null, t.tsInterfaceBody(fields));
}
function tsScalarForType(type, customScalarDefinition) {
    var alias;
    if (customScalarDefinition && customScalarDefinition.package) {
        alias = t.tsTypeReference(t.identifier(makeCustomScalarImportNameSafe(customScalarDefinition.name, type.name)));
    }
    else if (customScalarDefinition) {
        alias = t.tsTypeReference(t.identifier(customScalarDefinition.name));
    }
    else {
        alias = t.tsStringKeyword();
    }
    return t.tsTypeAliasDeclaration(t.identifier(type.name), null, alias);
}
function makeCustomScalarImportNameSafe(importName, typeName) {
    return "__" + typeName + "__" + importName;
}
function tsEnumForType(type, _a) {
    var enumFormat = _a.enumFormat;
    return t.tsEnumDeclaration(t.identifier(type.name), type
        .getValues()
        .map(function (value) {
        return t.tsEnumMember(t.identifier(enumMemberName(value.name, enumFormat)), t.stringLiteral(value.name));
    }));
}
function enumMemberName(name, format) {
    switch (format) {
        case types_1.EnumFormat.CamelCase:
            return change_case_1.camelCase(name);
        case types_1.EnumFormat.PascalCase:
            return change_case_1.pascalCase(name);
        case types_1.EnumFormat.SnakeCase:
            return change_case_1.snakeCase(name);
        case types_1.EnumFormat.ScreamingSnakeCase:
            return change_case_1.snakeCase(name).toUpperCase();
        default:
            return name;
    }
}
